searchNodes=[{"doc":"","ref":"Yildun.html","title":"Yildun","type":"module"},{"doc":"This module provides some boilerplate code for a better support of structs while using the MongoDB driver : automatic load and dump function reflection functions type specification support for embedding one and many structs support for after load function support for before dump function support for id generation support for default values support for derived values When using the MongoDB driver only maps and keyword lists are used to represent documents. If you would prefer to use structs instead of the maps to give the document a stronger meaning or to emphasize its importance, you have to create a defstruct and fill it from the map manually: defmodule Label do defstruct name : &quot;warning&quot; , color : &quot;red&quot; end iex&gt; label_map = Mongo . find_one ( :mongo , &quot;labels&quot; , %{ } ) %{ &quot;name&quot; =&gt; &quot;warning&quot; , &quot;color&quot; =&gt; &quot;red&quot; } iex&gt; label = % Label { name : label_map [ &quot;name&quot; ] , color : label_map [ &quot;color&quot; ] } We have defined a module Label as defstruct , then we get the first label document the collection labels . The function find_one returns a map. We convert the map manually and get the desired struct. If we want to save a new structure, we have to do the reverse. We convert the struct into a map: iex&gt; label = % Label { } iex&gt; label_map = %{ &quot;name&quot; =&gt; label . name , &quot;color&quot; =&gt; label . color } iex&gt; { :ok , _ } = Mongo . insert_one ( :mongo , &quot;labels&quot; , label_map ) Alternatively, you can also remove the __struct__ key from label . The MongoDB driver automatically converts the atom keys into strings. iex&gt; Map . drop ( label , [ :__struct__ ] ) %{ color : :red , name : &quot;warning&quot; } If you use nested structures, the work becomes a bit more complex. In this case, you have to use the inner structures convert manually, too. If you take a closer look at the necessary work, two basic functions can be derived: load Conversion of the map into a struct. dump Conversion of the struct into a map. This module provides the necessary macros to automate this boilerplate code. The above example can be rewritten as follows: defmodule Label do use Collection document do attribute :name , String . t ( ) , default : &quot;warning&quot; attribute :color , String . t ( ) , default : :red end end This results in the following module: defmodule Label do defstruct [ name : &quot;warning&quot; , color : &quot;red&quot; ] @type t ( ) :: % Label { String . t ( ) , String . t ( ) } def new ( ) ... def load ( map ) ... def dump ( % Label { } ) ... def __collection__ ( :attributes ) ... def __collection__ ( :types ) ... def __collection__ ( :collection ) ... def __collection__ ( :id ) ... end You can now create new structs with the default values and use the conversion functions between maps and structs: iex(1)&gt; x = Label . new ( ) % Label { color : :red , name : &quot;warning&quot; } iex(2)&gt; m = Label . dump ( x ) %{ color : :red , name : &quot;warning&quot; } iex(3)&gt; Label . load ( m , true ) % Label { color : :red , name : &quot;warning&quot; } The load/2 function distinguishes between keys of type binarys load(map, false) and keys of type atoms load(map, true) . The default is load(map, false) : iex(1)&gt; m = %{ &quot;color&quot; =&gt; :red , &quot;name&quot; =&gt; &quot;warning&quot; } iex(2)&gt; Label . load ( m ) % Label { color : :red , name : &quot;warning&quot; } If you would now expect atoms as keys, the result of the conversion is not correct in this case: iex(3)&gt; Label . load ( m , true ) % Label { color : nil , name : nil } The background is that MongoDB always returns binarys as keys and structs use atoms as keys. Default and derived values Attributes have two options: default: a value or a function, which is called, when a new struct is created derived: true to indicate, that is attribute should not be saved to the database If you call new/0 a new struct is returned filled with the default values. In case of a function the function is called to use the return value as default. attribute : created , DateTime . t ( ) , &amp; DateTime . utc_now / 0 If you mark an attribute as a derived attribute ( derived: true ) then the dump function will remove the attributes from the struct automatically for you, so these kind of attributes won't be saved in the database. attribute :id , String . t ( ) , derived : true Collections In MongoDB, documents are written in collections. We can use the collection/2 macro to create a collection: defmodule Card do use Collection @collection nil collection &quot;cards&quot; do attribute :title , String . t ( ) , &quot;new title&quot; end end The collection/2 macro creates a collection that is basically similar to a document, where an attribute for the ID is added automatically. Additionally the attribute @collection is assigned and can be used as a constant in other functions. In the example above we only suppress a warning of the editor by @collection . The macro creates the following expression: @collection &quot;cards&quot; . By default, the following attribute is created for the ID: { :_id , BSON.ObjectId . t ( ) , &amp; Mongo . object_id / 0 } where the default value is created via the function &amp;Mongo.object_id/0 when calling new/0 : iex&gt; Card . new ( ) % Card { _id : # BSON.ObjectId &lt; 5 ec3d04a306a5f296448a695 &gt; , title : &quot;new title&quot; } Two additional reflection features are also provided: iex&gt; Card . __collection__ ( :id ) :_id iex(3)&gt; Card . __collection__ ( :collection ) &quot;cards&quot; MongoDB example We define the following collection: defmodule Card do use Collection @collection nil ## keeps the editor happy @id nil collection &quot;cards&quot; do attribute :title , String . t ( ) , default : &quot;new title&quot; end def insert_one ( % Card { } = card ) do with map &lt;- dump ( card ) , { :ok , _ } &lt;- Mongo . insert_one ( :mongo , @collection , map ) do :ok end end def find_one ( id ) do :mongo |&gt; Mongo . find_one ( @collection , %{ @id =&gt; id } ) |&gt; load ( ) end end Then we can call the functions insert_one and find_one . Thereby we always use the defined structs as parameters or get the struct as result: iex(1)&gt; card = Card . new ( ) % Card { _id : # BSON.ObjectId &lt; 5 ec3ed0d306a5f377943c23c &gt; , title : &quot;new title&quot; } iex(6)&gt; Card . insert_one ( card ) :ok iex(2)&gt; Card . find_one ( card . _id ) % XCard { _id : # BSON.ObjectId &lt; 5 ec3ecbf306a5f3779a5edaa &gt; , title : &quot;new title&quot; } Id generator In MongoDB it is common to use the attribute _id as id. The value is uses an ObjectId generated by the mongodb driver. This behavior can be specified by the module attribute @id_generator when using collection . The default setting is { :_id , BSON.ObjectId . t ( ) , &amp; Mongo . object_id / 0 } Now you can overwrite this tuple {name, type, function} as you like: @id_generator false # no ID creation @id_generator { id , String . t , &amp; IDGenerator . next ( ) / 0 } # customized name and generator @id_generator nil # use default: {:_id, BSON.ObjectId.t(), &amp;Mongo.object_id/0} Embedded documents Until now we had only shown simple attributes. It will only be interesting when we embed other structs. With the macros embeds_one/3 and embeds_many/3 , structs can be added to the attributes: Example embeds_one defmodule Label do use Collection document do attribute :name , String . t ( ) , default : &quot;warning&quot; attribute :color , String . t ( ) , default : :red end end defmodule Card do use Collection collection &quot;cards&quot; do attribute :title , String . t ( ) attribute :list , BSON.ObjectId . t ( ) attribute :created , DateString . t ( ) , default : &amp; DateTime . utc_now / 0 attribute :modified , DateString . t ( ) , default : &amp; DateTime . utc_now / 0 embeds_one :label , Label , default : &amp; Label . new / 0 end end If we now call new/0 , we get the following structure: iex(1)&gt; Card . new ( ) % Card { _id : # BSON.ObjectId &lt; 5 ec3f0f0306a5f3aa5418a24 &gt; , created : ~U[2020-05-19 14:45:04.141044Z] , label : % Label { color : :red , name : &quot;warning&quot; } , list : nil , modified : ~U[2020-05-19 14:45:04.141033Z] , title : nil } after_load/1 and before_dump/1 macros Sometimes you may want to perform post-processing after loading the data set, for example to create derived attributes. Conversely, before saving, you may want to drop the derived attributes so that they are not saved to the database. For this reason there are two macros after_load/1 and before_dump/1 . You can specify functions that are called after the load/0 or before the dump : Example embeds_many defmodule Board do use Collection collection &quot;boards&quot; do attribute :id , String . t ( ) ## derived attribute attribute :title , String . t ( ) attribute :created , DateString . t ( ) , default : &amp; DateTime . utc_now / 0 attribute :modified , DateString . t ( ) , default : &amp; DateTime . utc_now / 0 embeds_many :lists , BoardList after_load &amp; Board . after_load / 1 before_dump &amp; Board . before_dump / 1 end def after_load ( % Board { _id : id } = board ) do % Board { board | id : BSON.ObjectId . encode! ( id ) } end def before_dump ( board ) do % Board { board | id : nil } end def new ( title ) do new ( ) |&gt; Map . put ( :title , title ) |&gt; Map . put ( :lists , [ ] ) |&gt; after_load ( ) end def store ( board ) do with map &lt;- dump ( board ) , { :ok , _ } &lt;- Mongo . insert_one ( :mongo , @collection , map ) do :ok end end def fetch ( id ) do :mongo |&gt; Mongo . find_one ( @collection , %{ @id =&gt; id } ) |&gt; load ( ) end end In this example the attribute id is derived from the actual ID and stored as a binary. This attribute is often used and therefore we want to save the conversion of the ID. To avoid storing the derived attribute id , the before_dump/1 function is called, which removes the id from the struct: iex(1)&gt; board = Board . new ( &quot;Vega&quot; ) % Board { _id : # BSON.ObjectId &lt; 5 ec3f802306a5f3ee3b71cf2 &gt; , created : ~U[2020-05-19 15:15:14.374556Z] , id : &quot;5ec3f802306a5f3ee3b71cf2&quot; , lists : [ ] , modified : ~U[2020-05-19 15:15:14.374549Z] , title : &quot;Vega&quot; } iex(2)&gt; Board . store ( board ) :ok iex(3)&gt; Board . fetch ( board . _id ) % Board { _id : # BSON.ObjectId &lt; 5 ec3f802306a5f3ee3b71cf2 &gt; , created : ~U[2020-05-19 15:15:14.374Z] , id : &quot;5ec3f802306a5f3ee3b71cf2&quot; , lists : [ ] , modified : ~U[2020-05-19 15:15:14.374Z] , title : &quot;Vega&quot; } If we call the document in the Mongo shell, we see that the attribute id was not stored there: &gt; db . boards . findOne ( { &quot;_id&quot; : ObjectId ( &quot;5ec3f802306a5f3ee3b71cf2&quot; ) } ) { &quot;_id&quot; : ObjectId ( &quot;5ec3f802306a5f3ee3b71cf2&quot; ) , &quot;created&quot; : ISODate ( &quot;2020-05-19T15:15:14.374Z&quot; ) , &quot;lists&quot; : [ ] , &quot;modified&quot; : ISODate ( &quot;2020-05-19T15:15:14.374Z&quot; ) , &quot;title&quot; : &quot;Vega&quot; }","ref":"Yildun.Collection.html","title":"Yildun.Collection","type":"module"},{"doc":"Adds the attribute to the attributes list.","ref":"Yildun.Collection.html#__attribute__/4","title":"Yildun.Collection.__attribute__/4","type":"function"},{"doc":"Adds the struct to the embeds_many list.","ref":"Yildun.Collection.html#__embeds_many__/5","title":"Yildun.Collection.__embeds_many__/5","type":"function"},{"doc":"Adds the struct to the embeds_one list.","ref":"Yildun.Collection.html#__embeds_one__/4","title":"Yildun.Collection.__embeds_one__/4","type":"function"},{"doc":"Inserts the specified @id_generator to the list of attributes. Calls add_id/3 .","ref":"Yildun.Collection.html#__id__/2","title":"Yildun.Collection.__id__/2","type":"macro"},{"doc":"Inserts boilercode for the @type attribute.","ref":"Yildun.Collection.html#__type__/1","title":"Yildun.Collection.__type__/1","type":"macro"},{"doc":"Inserts the specified @id_generator to the list of attributes.","ref":"Yildun.Collection.html#add_id/3","title":"Yildun.Collection.add_id/3","type":"function"},{"doc":"Defines the after_load/1 function.","ref":"Yildun.Collection.html#after_load/1","title":"Yildun.Collection.after_load/1","type":"macro"},{"doc":"Adds the attribute to the attributes list. It call __attribute__/4 function.","ref":"Yildun.Collection.html#attribute/3","title":"Yildun.Collection.attribute/3","type":"macro"},{"doc":"Defines the before_dump/1 function.","ref":"Yildun.Collection.html#before_dump/1","title":"Yildun.Collection.before_dump/1","type":"macro"},{"doc":"Defines a struct as a collection with id generator and a collection. Inside a collection block, each attribute is defined through the attribute/3 macro.","ref":"Yildun.Collection.html#collection/2","title":"Yildun.Collection.collection/2","type":"macro"},{"doc":"Defines a struct as a document without id generator and a collection. These documents are used to be embedded within collection structs. Inside a document block, each attribute is defined through the attribute/3 macro.","ref":"Yildun.Collection.html#document/1","title":"Yildun.Collection.document/1","type":"macro"},{"doc":"","ref":"Yildun.Collection.html#dump/1","title":"Yildun.Collection.dump/1","type":"function"},{"doc":"","ref":"Yildun.Collection.html#dump/2","title":"Yildun.Collection.dump/2","type":"function"},{"doc":"Adds the struct to the embeds_many list. Calls __embeds_many__","ref":"Yildun.Collection.html#embeds_many/3","title":"Yildun.Collection.embeds_many/3","type":"macro"},{"doc":"Adds the struct to the embeds_one list. Calls __embeds_one__","ref":"Yildun.Collection.html#embeds_one/3","title":"Yildun.Collection.embeds_one/3","type":"macro"},{"doc":"","ref":"Yildun.Collection.html#filter_nils/1","title":"Yildun.Collection.filter_nils/1","type":"function"},{"doc":"Returns true, if the Module has the dump/1 function.","ref":"Yildun.Collection.html#has_dump_function?/1","title":"Yildun.Collection.has_dump_function?/1","type":"function"},{"doc":"Returns true, if the Module has the load/1 function.","ref":"Yildun.Collection.html#has_load_function?/1","title":"Yildun.Collection.has_load_function?/1","type":"function"},{"doc":"Returns the default arguments for the struct. They are used to provide the default values in the new/0 call.","ref":"Yildun.Collection.html#struct_args/1","title":"Yildun.Collection.struct_args/1","type":"function"},{"doc":"","ref":"Yildun.Collection.html#struct_args/2","title":"Yildun.Collection.struct_args/2","type":"function"},{"doc":"Yildun Yildun is a white-hued star in the northern circumpolar constellation of Ursa Minor, forming the second star in the bear's tail. And it is a very small library to support structs while using the MongoDB driver for Elixir .","ref":"readme.html","title":"Yildun","type":"extras"},{"doc":"If available in Hex , the package can be installed by adding yildun to your list of dependencies in mix.exs : def deps do [ { :yildun , &quot;~&gt; 0.1.0&quot; } ] end Documentation can be generated with ExDoc and published on HexDocs . Once published, the docs can be found at https://hexdocs.pm/yildun .","ref":"readme.html#installation","title":"Yildun - Installation","type":"extras"}]